#include "x_core.h"
#include <core/scancodes.h>
#include <core/x_scancodes.h>
#include <avr/pgmspace.h>
#include <stdio.h>
#include <core/merlin_hw.h>
#include <core/x_helpers.h>

#include <config/settings.h>

#include <core/macros/mouse.h>
#include <core/macros/debug.h>

#define READ_SC_CODE(n) eeprom_read_byte(x_keyboard_table + n)


x_keyboard_data_t x_keyboard_data =
{
	.n_keys = 0,
	.mod_byte = 0,
	.keys = {0},
	.keyboard_changed = true,
};

x_macro_status_t x_macro_status =
{
	.modifiers_active = true,
	X_MOUSE_ACTIV_INIT //Use defines to avoid many #IF ... #ENDIF
	.debug_active = false,
};

x_macro_toggle_keys_t x_macro_toggle_keys EEMEM =
{
	.modifiers = X_SC_Q,
	X_MOUSE_TGL_KEY_INIT
	.debug = X_SC_D,
};


void x_core_set_macro(uint8_t *source_table, uint8_t size, uint8_t macro_value)
{
	uint8_t code, i;
	for(i = 0; i < size; i++)
	{
		code = eeprom_read_byte(source_table + i);
		eeprom_update_byte(x_keyboard_table + code, macro_value);
		
	}
}

void x_core_restore_macro(uint8_t *source_table, uint8_t *backup_table, uint8_t size)
{
	uint8_t code, sc_code, i;
	for(i = 0; i < size; i++)
	{
		code = eeprom_read_byte(source_table + i);
		sc_code = eeprom_read_byte(backup_table + i);
		//it is not a special keycode
		if(sc_code <= HID_KEYBOARD_SC_MEDIA_CALCULATOR)
		{
			eeprom_update_byte(x_keyboard_table + code, sc_code);
		}
	}
}

void x_core_backup_macro(uint8_t *source_table, uint8_t *backup_table, uint8_t size)
{
	uint8_t code, sc_code, i;
	for(i = 0; i < size; i++)
	{
		code = eeprom_read_byte(source_table + i);
		sc_code = eeprom_read_byte(x_keyboard_table + code);
		//it is not a special keycode
		if(sc_code <= HID_KEYBOARD_SC_MEDIA_CALCULATOR)
		{
			printf_debug("backup %d\n\r", sc_code);
			eeprom_update_byte(backup_table + i, sc_code);
		}
	}
}

void x_core_toggle_macro(uint8_t *source_table, uint8_t *backup_table, uint8_t size, uint8_t macro_value, bool *activ)
{
	if(*activ)
	{
		printf_debug("disable\n");
		x_core_restore_macro(source_table, backup_table, size);
		*activ = false;
	}
	else
	{
		printf_debug("enable\n");
		x_core_set_macro(source_table, size, macro_value);
		*activ = true;
	}
}

void x_core_init(void)
{
	x_core_backup_macro((void *)&x_modifiers, (void *)&x_modifiers_backup, sizeof(x_modifiers_t));
	X_MOUSE_BACKUP();
	x_core_set_macro((void*)&x_modifiers, sizeof(x_modifiers_t), HID_SPECIAL_MODIFIER);
}

void clear_hid_actions(void)
{
	x_keyboard_data.n_keys = 0;
	x_keyboard_data.mod_byte = 0;
	x_ir_reader.keyboard_changed = true;
	
  X_MOUSE_RESET();
}

uint8_t read_modifier(uint8_t x_code)
{
	uint8_t i, tmp;
	uint8_t ret = 1;
	for(i = 0; i < sizeof(x_modifiers_t); i++)
	{
		tmp = eeprom_read_byte(((uint8_t *)&x_modifiers) + i);
		if(tmp == x_code)
		{
			return ret;
		}
		ret <<= 1;
	}
	printf_debug("mod %d not found\n\r", x_code);
	return 0;
}

bool x_core_eval_macro_key(uint8_t *x_codes, uint8_t count)
{
	if(count != 2)
	{
		return false;
	}
	
	uint8_t code;
	
	if(READ_SC_CODE(x_codes[0]) == HID_SPECIAL_MACRO_KEY)
	{
		code = x_codes[1];
	}
	else if(READ_SC_CODE(x_codes[1]) == HID_SPECIAL_MACRO_KEY)
	{
		code = x_codes[0];
	}
	else
	{
		return false;
	}
	
	if(code == eeprom_read_byte(&x_macro_toggle_keys.modifiers))
	{
		printf_debug("toggle modifiers\n");
		x_core_toggle_macro((uint8_t *)&x_modifiers, (uint8_t *)&x_modifiers_backup, sizeof(x_modifiers_t), HID_SPECIAL_MODIFIER, &x_macro_status.modifiers_active);
		x_ir_reader.ignore_keys_until_next_timeout = true;
		return true;
	}
	
#if X_MOUSE > 0
	if(code == eeprom_read_byte(&x_macro_toggle_keys.mouse))
	{
		X_MOUSE_TGL();
		x_ir_reader.ignore_keys_until_next_timeout = true;
		return true;
	}
#endif
	
	if(code == eeprom_read_byte(&x_macro_toggle_keys.debug))
	{
		x_macro_status.debug_active = !x_macro_status.debug_active;
		x_ir_reader.ignore_keys_until_next_timeout = true;
		return true;
	}
	
	return false;

}

void translate(uint8_t* x_codes, uint8_t count)
{
	if(count == 0)
	{
		clear_hid_actions();
		return;
	}
	uint8_t i, j, sc_code, x_code;
	uint8_t modifier = 0;
	bool changed = false;
	j = 0;
#if X_MOUSE > 0
	bool mouse_moved = false;
#endif
	for(i = 0; i < count; i++)
	{
		x_code = x_codes[i];
		if(x_code >= 0x80)
		{
			printf_debug("code to high: %d\n\r", x_code);
			return;
		}
		sc_code = eeprom_read_byte(x_keyboard_table + x_code);
		printf_debug("x: %d -> %d\n\r", x_code, sc_code);
		if(sc_code > HID_KEYBOARD_SC_MEDIA_CALCULATOR)
		{
			switch(sc_code)
			{
				case HID_SPECIAL_MODIFIER:
					modifier |= read_modifier(x_codes[i]);
					continue;
				case HID_SPECIAL_MACRO_KEY:
					x_core_eval_macro_key(x_codes, count);
					clear_hid_actions();
					return;
#if X_MOUSE > 0
				case HID_SPECIAL_MOUSE:
					mouse_moved |= eval_mouse(x_code);
					continue;
#endif
			}
		}
		else
		{
			changed = true;
			x_keyboard_data.keys[j] = sc_code;
			j++;
		}
	}
	
#if X_MOUSE > 0
	if(mouse_moved)
	{
		if(x_mouse_info.mouse_jump < MERLIN_MOUSE_MAX_SPEED)
		{
			x_mouse_info.mouse_jump++;
		}
	}
	else
	{
		x_mouse_info.mouse_jump = 1;
	}
#endif

	x_keyboard_data.mod_byte = modifier;
	x_keyboard_data.n_keys = j;

	x_keyboard_data.keyboard_changed |= changed;
	return;
}


x_modifiers_t x_modifiers EEMEM =
{
	.modifier_leftctrl    = X_SC_LEFT_CONTROL,
	.modifier_leftshift   = X_SC_LEFT_SHIFT,
	.modifier_leftalt     = X_SC_LEFT_ALT,
	.modifier_leftgui     = X_SC_LEFT_GUI,
	.modifier_rightctrl   = X_SC_RIGHT_CONTROL,
	.modifier_rightshift  = X_SC_RIGHT_SHIFT,
	.modifier_rightalt    = X_SC_RIGHT_ALT,
	.modifier_rightgui    = X_SC_RIGHT_GUI,
};


x_modifiers_t x_modifiers_backup EEMEM =
{
	0,
	0,
	0,
	0,
	0,
	0,
};


uint8_t x_keyboard_table[] EEMEM =
{
	/*0x00*/ 0,
	/*0x01*/ HID_SPECIAL_MACRO_KEY,
	/*0x02*/ HID_KEYBOARD_SC_ENTER,
	/*0x03*/ HID_KEYBOARD_SC_ESCAPE,
	/*0x04*/ HID_KEYBOARD_SC_BACKSPACE,
	/*0x05*/ HID_KEYBOARD_SC_TAB,
	/*0x06*/ HID_KEYBOARD_SC_SPACE,
	/*0x07*/ HID_KEYBOARD_SC_CAPS_LOCK,
	/*0x08*/ HID_KEYBOARD_SC_F1,
	/*0x09*/ HID_KEYBOARD_SC_F2,
	/*0x0A*/ HID_KEYBOARD_SC_F3,
	/*0x0B*/ HID_KEYBOARD_SC_F4,
	/*0x0C*/ HID_KEYBOARD_SC_F5,
	/*0x0D*/ HID_KEYBOARD_SC_F6,
	/*0x0E*/ HID_KEYBOARD_SC_F7,
	/*0x0F*/ HID_KEYBOARD_SC_F8,
	/*0x10*/ HID_KEYBOARD_SC_F9,
	/*0x11*/ HID_KEYBOARD_SC_F10,
	/*0x12*/ HID_KEYBOARD_SC_F11,
	/*0x13*/ HID_KEYBOARD_SC_F12,
	/*0x14*/ HID_KEYBOARD_SC_1_AND_EXCLAMATION,
	/*0x15*/ HID_KEYBOARD_SC_2_AND_AT,
	/*0x16*/ HID_KEYBOARD_SC_3_AND_HASHMARK,
	/*0x17*/ HID_KEYBOARD_SC_4_AND_DOLLAR,
	/*0x18*/ HID_KEYBOARD_SC_5_AND_PERCENTAGE,
	/*0x19*/ HID_KEYBOARD_SC_6_AND_CARET,
	/*0x1A*/ HID_KEYBOARD_SC_7_AND_AMPERSAND,
	/*0x1B*/ HID_KEYBOARD_SC_8_AND_ASTERISK,
	/*0x1C*/ HID_KEYBOARD_SC_9_AND_OPENING_PARENTHESIS,
	/*0x1D*/ HID_KEYBOARD_SC_0_AND_CLOSING_PARENTHESIS,
	/*0x1E*/ HID_KEYBOARD_SC_A,
	/*0x1F*/ HID_KEYBOARD_SC_B,
	/*0x20*/ HID_KEYBOARD_SC_C,
	/*0x21*/ HID_KEYBOARD_SC_D,
	/*0x22*/ HID_KEYBOARD_SC_E,
	/*0x23*/ HID_KEYBOARD_SC_F,
	/*0x24*/ HID_KEYBOARD_SC_G,
	/*0x25*/ HID_KEYBOARD_SC_H,
	/*0x26*/ HID_KEYBOARD_SC_I,
	/*0x27*/ HID_KEYBOARD_SC_J,
	/*0x28*/ HID_KEYBOARD_SC_K,
	/*0x29*/ HID_KEYBOARD_SC_L,
	/*0x2A*/ HID_KEYBOARD_SC_M,
	/*0x2B*/ HID_KEYBOARD_SC_N,
	/*0x2C*/ HID_KEYBOARD_SC_O,
	/*0x2D*/ HID_KEYBOARD_SC_P,
	/*0x2E*/ HID_KEYBOARD_SC_Q,
	/*0x2F*/ HID_KEYBOARD_SC_R,
	/*0x30*/ HID_KEYBOARD_SC_S,
	/*0x31*/ HID_KEYBOARD_SC_T,
	/*0x32*/ HID_KEYBOARD_SC_U,
	/*0x33*/ HID_KEYBOARD_SC_V,
	/*0x34*/ HID_KEYBOARD_SC_W,
	/*0x35*/ HID_KEYBOARD_SC_X,
	/*0x36*/ HID_KEYBOARD_SC_Y,
	/*0x37*/ HID_KEYBOARD_SC_Z,
	/*0x38*/ HID_KEYBOARD_SC_MINUS_AND_UNDERSCORE,
	/*0x39*/ HID_KEYBOARD_SC_EQUAL_AND_PLUS,
	/*0x3A*/ HID_KEYBOARD_SC_OPENING_BRACKET_AND_OPENING_BRACE,
	/*0x3B*/ HID_KEYBOARD_SC_CLOSING_BRACKET_AND_CLOSING_BRACE,
	/*0x3C*/ HID_KEYBOARD_SC_BACKSLASH_AND_PIPE,
	/*0x3D*/ HID_KEYBOARD_SC_NON_US_HASHMARK_AND_TILDE,
	/*0x3E*/ HID_KEYBOARD_SC_SEMICOLON_AND_COLON,
	/*0x3F*/ HID_KEYBOARD_SC_APOSTROPHE_AND_QUOTE,
	/*0x40*/ HID_KEYBOARD_SC_GRAVE_ACCENT_AND_TILDE,
	/*0x41*/ HID_KEYBOARD_SC_COMMA_AND_LESS_THAN_SIGN,
	/*0x42*/ HID_KEYBOARD_SC_DOT_AND_GREATER_THAN_SIGN,
	/*0x43*/ HID_KEYBOARD_SC_SLASH_AND_QUESTION_MARK,
	/*0x44*/ HID_KEYBOARD_SC_PRINT_SCREEN,
	/*0x45*/ HID_KEYBOARD_SC_SCROLL_LOCK,
	/*0x46*/ HID_KEYBOARD_SC_PAUSE,
	/*0x47*/ HID_KEYBOARD_SC_INSERT,
	/*0x48*/ HID_KEYBOARD_SC_HOME,
	/*0x49*/ HID_KEYBOARD_SC_PAGE_UP,
	/*0x4A*/ HID_KEYBOARD_SC_DELETE,
	/*0x4B*/ HID_KEYBOARD_SC_END,
	/*0x4C*/ HID_KEYBOARD_SC_PAGE_DOWN,
	/*0x4D*/ HID_KEYBOARD_SC_RIGHT_ARROW,
	/*0x4E*/ HID_KEYBOARD_SC_LEFT_ARROW,
	/*0x4F*/ HID_KEYBOARD_SC_DOWN_ARROW,
	/*0x50*/ HID_KEYBOARD_SC_UP_ARROW,
	/*0x51*/ HID_KEYBOARD_SC_POWER,
	/*0x52*/ HID_KEYBOARD_SC_MUTE,
	/*0x53*/ HID_KEYBOARD_SC_LEFT_CONTROL,
 	/*0x54*/ HID_KEYBOARD_SC_LEFT_SHIFT,
	/*0x55*/ HID_KEYBOARD_SC_LEFT_ALT,
	/*0x56*/ HID_KEYBOARD_SC_LEFT_GUI,
	/*0x57*/ HID_KEYBOARD_SC_RIGHT_CONTROL,
	/*0x58*/ HID_KEYBOARD_SC_RIGHT_SHIFT,
	/*0x59*/ HID_KEYBOARD_SC_RIGHT_ALT,
	/*0x5A*/ HID_KEYBOARD_SC_RIGHT_GUI,
	/*0x5B*/ HID_KEYBOARD_SC_MEDIA_PLAY,
	/*0x5C*/ HID_KEYBOARD_SC_MEDIA_STOP,
	/*0x5D*/ HID_KEYBOARD_SC_MEDIA_PREVIOUS_TRACK,
	/*0x5E*/ HID_KEYBOARD_SC_MEDIA_NEXT_TRACK,
	/*0x5F*/ HID_KEYBOARD_SC_MEDIA_VOLUME_UP,
	/*0x60*/ HID_KEYBOARD_SC_MEDIA_VOLUME_DOWN,
	/*0x61*/ HID_KEYBOARD_SC_MEDIA_MUTE,
	/*0x62*/ HID_KEYBOARD_SC_MEDIA_BACKWARD,
	/*0x63*/ HID_KEYBOARD_SC_MEDIA_FORWARD,
	/*0x64*/ HID_KEYBOARD_SC_NUM_LOCK,
	/*0x65*/ HID_KEYBOARD_SC_KEYPAD_SLASH,
	/*0x66*/ HID_KEYBOARD_SC_KEYPAD_ASTERISK,
	/*0x67*/ HID_KEYBOARD_SC_KEYPAD_MINUS,
	/*0x68*/ HID_KEYBOARD_SC_KEYPAD_PLUS,
	/*0x69*/ HID_KEYBOARD_SC_KEYPAD_ENTER,
	/*0x6A*/ HID_KEYBOARD_SC_KEYPAD_1_AND_END,
	/*0x6B*/ HID_KEYBOARD_SC_KEYPAD_2_AND_DOWN_ARROW,
	/*0x6C*/ HID_KEYBOARD_SC_KEYPAD_3_AND_PAGE_DOWN,
	/*0x6D*/ HID_KEYBOARD_SC_KEYPAD_4_AND_LEFT_ARROW,
	/*0x6E*/ HID_KEYBOARD_SC_KEYPAD_5,
	/*0x6F*/ HID_KEYBOARD_SC_KEYPAD_6_AND_RIGHT_ARROW,
	/*0x70*/ HID_KEYBOARD_SC_KEYPAD_7_AND_HOME,
	/*0x71*/ HID_KEYBOARD_SC_KEYPAD_8_AND_UP_ARROW,
	/*0x72*/ HID_KEYBOARD_SC_KEYPAD_9_AND_PAGE_UP,
	/*0x73*/ HID_KEYBOARD_SC_KEYPAD_0_AND_INSERT,
	/*0x74*/ HID_KEYBOARD_SC_KEYPAD_DOT_AND_DELETE,
	/*0x75*/ X_SC_SPECIAL_MAP1,
	/*0x76*/ X_SC_SPECIAL_MAP2,
	/*0x77*/ X_SC_SPECIAL_MAP3,
	/*0x78*/ X_SC_SPECIAL_MAP4,
	/*0x79*/ X_SC_SPECIAL_MAP5,
	/*0x7A*/ X_SC_SPECIAL_MAP6,
	/*0x7B*/ X_SC_SPECIAL_MAP7,
	/*0x7C*/ X_SC_SPECIAL_MAP8,
	/*0x7D*/ X_SC_SPECIAL_MAP9,
	/*0x7E*/ X_SC_SPECIAL_MAP10,
	/*0x7F*/ X_SC_SPECIAL_MAP11,
};




